"""Assigning default values."""

from flask_jeroboam import Jeroboam, Query

app = Jeroboam("Jeroboam Inbound Features App")
app.init_app()


@app.get("/implicit_location_with_default_value")
def get_implicit_argument_with_default(page: int = 1):
    return f"Received Page Argument is : {page}"


@app.get("/explicit_location_with_default_value")
def get_explicit_argument_with_default(page: int = Query(1)):
    return f"Received Page Argument is : {page}"


if __name__ == "__main__":
    app.run(host="localhost", port=5000)



def test_text_get(client):
    """Test GET /text"""
    response = client.get("/text")
    assert response.status_code == 200, response.text
    assert response.data == b"Hello World"


def test_nonexistent(client):
    """Test GET /nonexistent"""
    response = client.get("/nonexistent")
    assert response.status_code == 404, response.text
    assert response.json == {"message": "Not Found"}



"""OpenAPI auto-documentation - initializing Jeroboam app."""

from flask_jeroboam import Jeroboam

app = Jeroboam(__name__)
app.init_app()


if __name__ == "__main__":
    app.run(host="localhost", port=5000)



"""Custom responses for Flask-Jeroboam."""

from flask import Response


class JeroboamResponse(Response):
    """Base class for all Jeroboam responses."""

    default_status_code = 200


class JSONResponse(JeroboamResponse):
    """Subclassing Flask Response for JSON responses."""

    default_mimetype = "application/json"


class HTMLResponse(JeroboamResponse):
    """Subclassing Flask Response for HTML responses."""

    default_mimetype = "text/html"



"""Mixing implicit and explicit locations."""

from flask_jeroboam import Cookie, Jeroboam, Query

app = Jeroboam("Jeroboam Inbound Features App")
app.init_app()


@app.get("/explicit_location_is_query_string_and_cookie")
def get_explicit_arguments(page: int = Query(), username: str = Cookie()):
    return f"Received Page Argument is : {page}. Username is : {username}"


@app.get("/implicit_and_explicit")
def get_implicit_and_explicit_arguments(page: int, username: str = Cookie()):
    return f"Received Page Argument is : {page}. Username is : {username}"


if __name__ == "__main__":
    app.run(host="localhost", port=5000)



"""A Test Blueprint for testing Outbound Behavior.

The corresponding test can be found in tests/test_outbound.py
"""

from typing import Any

from flask_jeroboam import Blueprint
from flask_jeroboam.responses import JSONResponse
from flask_jeroboam.view_arguments.functions import Body
from tests.app_test.models.outbound import MyDataClass, SimpleModelOut, UserIn, UserOut

router = Blueprint("outbound_router", __name__, tags=["Outbound"])

valid_outbound_data: dict[str, Any] = {
    "total_count": 10,
    "items": ["Apple", "Banana"],
}


@router.route(
    "/methods/explicit_options",
    methods=["GET", "OPTIONS"],
    response_model=SimpleModelOut,
)
def get_with_explicit_verb_options():
    return valid_outbound_data


@router.route(
    "/methods/explicit_options_and_head",
    methods=["GET", "OPTIONS", "HEAD"],
    response_model=SimpleModelOut,
)
def get_with_explicit_verbs_options_and_head():
    return valid_outbound_data


@router.put("/verb/put/without_explicit_status_code", response_model=SimpleModelOut)
def put_http_verb():
    return valid_outbound_data


@router.patch("/verb/patch/without_explicit_status_code", response_model=SimpleModelOut)
def patch_http_verb():
    return valid_outbound_data


@router.get("/response_model/no_response_model")
def no_response_model():
    return "Don't have a response model"


@router.get("/response_model/infered_from_return_annotation")
def response_model_is_infered_from_return_annotation() -> SimpleModelOut:
    return SimpleModelOut(total_count=10, items=["Apple", "Banana"])


@router.get(
    "/response_model/configuration_over_inference", response_model=SimpleModelOut
)
def configuration_over_inference() -> dict:
    return valid_outbound_data


@router.get("/response_model/turned_off", response_model=None)
def response_model_inference_is_turned_off() -> SimpleModelOut:
    return SimpleModelOut(total_count=10, items=["Apple", "Banana"])


@router.get("/return_type/dict", response_model=SimpleModelOut)
def view_function_returns_a_dict() -> dict:
    return valid_outbound_data


@router.get("/return_type/list", response_model=list[SimpleModelOut])
def view_function_returns_a_list():
    return [valid_outbound_data, valid_outbound_data]


@router.get("/return_type/base_model", response_model=SimpleModelOut)
def view_function_returns_a_base_model():
    return SimpleModelOut(total_count=10, items=["Apple", "Banana"])


@router.get("/return_type/response", response_model=SimpleModelOut)
def view_function_returns_a_response():
    return JSONResponse(
        SimpleModelOut(total_count=10, items=["Apple", "Banana"]).model_dump_json()
    )


@router.get("/return_type/dataclass", response_model=SimpleModelOut)
def view_function_returns_a_dataclass():
    return MyDataClass(**valid_outbound_data)


@router.get("/return_type/not_valid", response_model=SimpleModelOut)
def view_function_returns_unvalid():
    return "not a list"


@router.get("/return_shape/with_headers", response_model=SimpleModelOut)
def view_function_returns_dict_and_headers():
    return valid_outbound_data, {"X-Test": "Test"}


@router.get("/return_shape/with_headers_and_status_code", response_model=SimpleModelOut)
def view_function_returns_dict_status_code_and_headers():
    return valid_outbound_data, 218, {"X-Test": "Test"}


@router.get("/return_shape/with_status_code", response_model=SimpleModelOut)
def view_function_returns_dict_and_status_code():
    return valid_outbound_data, 218


@router.get("/return_shape/wrong_tuple_length", response_model=SimpleModelOut)
def view_function_returns_wrong_tuple_length():
    return valid_outbound_data, 200, {"X-Test": "Test"}, "extra"


@router.get("/status_code/204_has_no_body/as_returned")
def returned_status_code_204_has_no_body_returned():
    return "Some Content that will be ignored", 204


@router.get(
    "/status_code/204_has_no_body/as_configured",
    status_code=204,
)
def configured_status_code_204_has_no_body():
    return "Some Content that will be ignored"


@router.post("/sensitive_data", response_model=UserOut)
def reponse_model_filters_data(sensitive_data: UserIn = Body()):
    return sensitive_data



"""Models for OpenAPI 3.0.0."""

from pydantic import Field

from flask_jeroboam.models import OutboundModel


class SwaggerContextOut(OutboundModel):
    """Swagger Outbound Model."""

    title: str | None = Field(
        default="Swagger UI", description="Title of the Swagger UI"
    )
    swagger_ui_url: str | None = Field(
        default="https://cdn.jsdelivr.net/npm/swagger-ui-dist@4",
        description="The Base URL for all Swagger Ui Assets",
    )
    current_swagger_ui_parameters: dict | None = Field(
        default={}, description="The current Swagger UI parameters"
    )



"""Helper functions for extracting values from request locations."""

from werkzeug.datastructures import Headers, MultiDict

from flask_jeroboam._utils import is_sequence_field
from flask_jeroboam.wrapper import current_app


def _extract_scalar(
    *,
    source: Headers | dict,
    name: str,
    alias: str,
    **_kwargs,
):
    """Extract a scalar value from a source."""
    return source.get(alias, source.get(name))


def _extract_sequence(
    *, source: MultiDict, name: str | None, alias: str | None, **_kwargs
) -> list:
    """Extract a Sequence value from a source."""
    _values = source.getlist(alias)
    if len(_values) == 0:
        _values = source.getlist(name)
    return _values


def _extract_sequence_with_key_transformer(
    *, source: MultiDict, name: str, alias: str, **_kwargs
):
    """Apply the key transformer to the source."""
    transformed_source = current_app.query_string_key_transformer(
        current_app, source.to_dict()
    )
    return _extract_scalar(source=transformed_source, name=name, alias=alias)


def _undirected_extraction(
    *,
    field,
    source,
    alias: str,
    name: str,
    has_key_transformer: bool,
    **_kwargs,
):
    if is_sequence_field(field):
        values = _extract_sequence(source=source, name=name, alias=alias)
        if len(values) == 0 and has_key_transformer:
            values = _extract_sequence_with_key_transformer(
                source=source, name=name, alias=alias
            )
    else:
        values = _extract_scalar(source=source, name=name, alias=alias)
    return values


def _extract_subfields(
    *,
    source: MultiDict,
    fields: dict,
    **_kwargs,
) -> dict:
    """Extract a Sequence from subfields."""
    has_key_transformer = (
        getattr(current_app, "query_string_key_transformer", False) is not None
    )
    result = {}
    for field_name, subfield in fields.items():
        value = _undirected_extraction(
            field=subfield,
            source=source,
            name=field_name,
            alias=subfield.alias,
            has_key_transformer=has_key_transformer,
        )
        if value is not None:
            result[field_name] = value
    return result



"""A Test Blueprint for testing Header Params.

The corresponding test can be found in tests/test_inbound/test_header
"""

from flask_jeroboam import Blueprint, Header

router = Blueprint("headers_params_router", __name__, tags=["Header"])


@router.get("/headers/int")
def get_header_as_plain_int(test_header: int = Header()):
    return {"header": test_header}


@router.get("/headers/str")
def get_header_as_plain_str(test_header: str = Header()):
    return {"header": test_header}



"""Flask-Jeroboam Blueprint.

Subclasses the Flask Blueprint.
Can be used as a drop in replacement of any Flask's Blueprint.
"""

import os

from flask import Blueprint as FlaskBlueprint

from flask_jeroboam.scaffold import JeroboamScaffoldOverRide

_sentinel = object()

class Blueprint(JeroboamScaffoldOverRide, FlaskBlueprint):  # type:ignore
    def __init__(
        self,
        name: str,
        import_name: str,
        static_folder: str | os.PathLike | None = None,
        static_url_path: str | None = None,
        template_folder: str | os.PathLike | None = None,
        url_prefix: str | None = None,
        subdomain: str | None = None,
        url_defaults: dict | None = None,
        root_path: str | None = None,
        cli_group: str | None = _sentinel,  # type: ignore
        tags: list[str] | None = None,
        include_in_openapi: bool = True,
    ) -> None: ...



"""Test OpenAPI infos."""


def test_info(client):
    """Test info endpoint."""
    response = client.get("/openapi.json")
    assert response.status_code == 200
    assert response.json["info"] == {
        "contact": {
            "email": "jc.bianic@gmail.com",
            "name": "Jean-Christophe Bianic",
        },
        "description": "This application is here to test Jeroboam OpenAPI features.",
        "license": {"name": "MIT", "url": "http://opensource.org/licenses/MIT"},
        "termsOfService": "http://example.com/terms/",
        "title": "Jeroboam Test App",
        "version": "0.1.0",
    }



"""Stub file for the adjacent functions module.

Note that the default setting behavior is implemented
in each __init__ methods.

Documentation on stubfiles can be found here:
https://mypy.readthedocs.io/en/stable/stubs.html
"""

from typing import Any

from pydantic_core import PydanticUndefined

def Path(
    default: Any = PydanticUndefined,
    *,
    alias: str | None = None,
    title: str | None = None,
    description: str | None = None,
    gt: float | None = None,
    ge: float | None = None,
    lt: float | None = None,
    le: float | None = None,
    min_length: int | None = None,
    max_length: int | None = None,
    pattern: str | None = None,
    example: Any = PydanticUndefined,
    examples: dict[str, Any] | None = None,
    deprecated: bool | None = None,
    include_in_schema: bool = True,
    **extra: Any,
) -> Any: ...
def Query(
    default: Any = PydanticUndefined,
    *,
    alias: str | None = None,
    title: str | None = None,
    description: str | None = None,
    gt: float | None = None,
    ge: float | None = None,
    lt: float | None = None,
    le: float | None = None,
    min_length: int | None = None,
    max_length: int | None = None,
    pattern: str | None = None,
    example: Any = PydanticUndefined,
    examples: dict[str, Any] | None = None,
    deprecated: bool | None = None,
    include_in_schema: bool = True,
    **extra: Any,
) -> Any: ...
def Header(
    default: Any = PydanticUndefined,
    *,
    alias: str | None = None,
    title: str | None = None,
    description: str | None = None,
    gt: float | None = None,
    ge: float | None = None,
    lt: float | None = None,
    le: float | None = None,
    min_length: int | None = None,
    max_length: int | None = None,
    pattern: str | None = None,
    example: Any = PydanticUndefined,
    examples: dict[str, Any] | None = None,
    deprecated: bool | None = None,
    include_in_schema: bool = True,
    convert_underscores: bool = True,
    **extra: Any,
) -> Any: ...
def Cookie(
    default: Any = PydanticUndefined,
    *,
    alias: str | None = None,
    title: str | None = None,
    description: str | None = None,
    gt: float | None = None,
    ge: float | None = None,
    lt: float | None = None,
    le: float | None = None,
    min_length: int | None = None,
    max_length: int | None = None,
    pattern: str | None = None,
    example: Any = PydanticUndefined,
    examples: dict[str, Any] | None = None,
    deprecated: bool | None = None,
    include_in_schema: bool = True,
    **extra: Any,
) -> Any: ...
def Body(
    default: Any = PydanticUndefined,
    *,
    alias: str | None = None,
    title: str | None = None,
    description: str | None = None,
    gt: float | None = None,
    ge: float | None = None,
    lt: float | None = None,
    le: float | None = None,
    min_length: int | None = None,
    max_length: int | None = None,
    pattern: str | None = None,
    example: Any = PydanticUndefined,
    examples: dict[str, Any] | None = None,
    deprecated: bool | None = None,
    include_in_schema: bool = True,
    required: bool = False,
    embed: bool = True,  # for body
    media_type: str = "application/json",
    **extra: Any,
) -> Any: ...
def Form(
    default: Any = PydanticUndefined,
    *,
    alias: str | None = None,
    title: str | None = None,
    description: str | None = None,
    gt: float | None = None,
    ge: float | None = None,
    lt: float | None = None,
    le: float | None = None,
    min_length: int | None = None,
    max_length: int | None = None,
    pattern: str | None = None,
    example: Any = PydanticUndefined,
    examples: dict[str, Any] | None = None,
    deprecated: bool | None = None,
    include_in_schema: bool = True,
    required: bool = False,
    embed: bool = True,
    media_type: str = "application/x-www-form-urlencoded",
    **extra: Any,
) -> Any: ...
def File(  # noqa: N802
    default: Any = PydanticUndefined,
    *,
    alias: str | None = None,
    title: str | None = None,
    description: str | None = None,
    gt: float | None = None,
    ge: float | None = None,
    lt: float | None = None,
    le: float | None = None,
    min_length: int | None = None,
    max_length: int | None = None,
    pattern: str | None = None,
    example: Any = PydanticUndefined,
    examples: dict[str, Any] | None = None,
    deprecated: bool | None = None,
    include_in_schema: bool = True,
    required: bool = False,
    embed: bool = False,
    media_type: str = "multipart/form-data",
    **extra: Any,
) -> Any: ...



"""Stub File for Jeroboam."""

from collections.abc import Callable
from typing import Any

from flask import Flask
from typing_extensions import TypeVar

from flask_jeroboam.openapi.models.openapi import OpenAPI
from flask_jeroboam.rule import JeroboamRule
from flask_jeroboam.scaffold import JeroboamScaffoldOverRide

R = TypeVar("R", bound=Any)
_sentinel = object()
current_app: Jeroboam

class Jeroboam(JeroboamScaffoldOverRide, Flask):  # type:ignore
    query_string_key_transformer: Callable
    openapi: OpenAPI
    def rules(self) -> list[JeroboamRule]: ...
    def init_app(self, app: Jeroboam | None = None) -> None: ...



"""Function to declare the Type of Parameters.

This functions are used to declare the parameters of the view functions.
By annotating the return value with Any, we make sure that the code editor
don't complain about assigning a default value of type ViewParameter
to a parameter that have been annotated with a pydantic-compatible type...
They're primary purprose is to trick the code editor as they only delegate
to actual ViewParameter instantiers.
Their signature are defined in adjacent file functions.pyi.

Credits: This module is a fork of FlaskAPI params_function module.
"""

from typing import Any

from flask_jeroboam.view_arguments.arguments import (
    BodyArgument,
    CookieArgument,
    FileArgument,
    FormArgument,
    HeaderArgument,
    PathArgument,
    QueryArgument,
)


def Path(  # noqa:N802
    *args: Any,
    **kwargs: Any,
) -> Any:
    """Declare A Path parameter."""
    return PathArgument(
        *args,
        **kwargs,
    )


def Query(  # noqa:N802
    *args: Any,
    **kwargs: Any,
) -> Any:
    """Declare A Query parameter."""
    return QueryArgument(
        *args,
        **kwargs,
    )


def Header(  # noqa:N802
    *args: Any,
    **kwargs: Any,
) -> Any:
    """Declare A Header parameter."""
    return HeaderArgument(
        *args,
        **kwargs,
    )


def Cookie(  # noqa:N802
    *args: Any,
    **kwargs: Any,
) -> Any:
    """Declare A Cookie parameter."""
    return CookieArgument(
        *args,
        **kwargs,
    )


def Body(  # noqa:N802
    *args: Any,
    **kwargs: Any,
) -> Any:
    """Declare A Body parameter."""
    return BodyArgument(
        *args,
        **kwargs,
    )


def Form(  # noqa: N802
    *args: Any,
    **kwargs: Any,
) -> Any:
    """Declare A Form parameter."""
    return FormArgument(
        *args,
        **kwargs,
    )


def File(  # noqa: N802
    *args: Any,
    **kwargs: Any,
) -> Any:
    """Declare A File parameter."""
    return FileArgument(
        *args,
        **kwargs,
    )



"""Model Helpers."""

from pydantic import BaseModel, ConfigDict


def snake_case_to_camel(string: str) -> str:
    """Convert snake_case_string to aCamelCaseString."""
    camel_case = string.title().replace("_", "")
    return camel_case[0].lower() + camel_case[1:]


class InboundModel(BaseModel):
    """Basic configuration for parsing Requests."""

    model_config = ConfigDict(
        alias_generator=snake_case_to_camel,
        populate_by_name=True,
    )


class OutboundModel(BaseModel):
    """Basic configuration for serializing Responses."""

    model_config = ConfigDict(
        alias_generator=snake_case_to_camel,
        populate_by_name=True,
    )

    def model_dump_json(self, **kwargs) -> str:
        """Serialize to JSON with camelCase keys by default."""
        kwargs.setdefault("by_alias", True)
        return super().model_dump_json(**kwargs)



"""A Test Blueprint for testing Misc Behavior."""

from flask_jeroboam import Blueprint
from flask_jeroboam.exceptions import RessourceNotFound, ServerError

router = Blueprint("misc_router", __name__, tags=["Misc"])


@router.get("/invalid_request")
def get_invalid_request(missing_param: int):
    return {"missing_param": missing_param}


@router.get("/ressource_not_found")
def get_ressource_not_found():
    raise RessourceNotFound(ressource_name="TestRessource", context=f"with id {id}")


@router.get("/generic_ressource")
def get_generic_ressource_not_found():
    raise RessourceNotFound(msg="My Message")


@router.get("/server_error")
def get_a_server_error():
    raise ServerError(msg="My Message", error=Exception(), trace="FakeTrace")


@router.delete("/delete")
def delete():
    return {}



"""Code Sample for Getting Started with Flask-Jeroboam."""

from pydantic import BaseModel, Field

from flask_jeroboam import Jeroboam

app = Jeroboam("Jeroboam Getting Started App")
app.init_app()


@app.get("/health")
def get_health():
    return {"status": "Ok"}


@app.get("/query_string_arguments")
def get_query_string_arguments(page: int = 1, per_page: int = 10):
    return {"page": page, "per_page": per_page}


class Item(BaseModel):
    name: str
    price: float
    count: int = Field(1)


@app.get("/item", response_model=Item)
def get_an_item():
    return {"name": "Bottle", "price": 5}


if __name__ == "__main__":
    app.run(host="localhost", port=5000)



"""A Test Blueprint for testing Form Params.

The corresponding test can be found in tests/test_inbound/test_form
"""

from flask_jeroboam import Blueprint, Form
from tests.app_test.models.inbound import SimpleModelIn

router = Blueprint("form_params_router", __name__, tags=["Form"])


@router.post("/form/base_model")
def post_base_model_in_form(form_payload: SimpleModelIn = Form(embed=False)):
    """POST Form Parameter as pydantic BaseModel."""
    return form_payload.model_dump_json()


@router.post("/form/single_primitive")
def post_single_primitive_in_form(name: str = Form("John")):
    """POST a single primitive Form parameter."""
    return {"name": name}


@router.post("/form/primitives_in_form")
def post_primitives_in_form(name: str = Form("John"), age: int = Form(34)):
    """POST Form Parameter as pydantic BaseModel."""
    return {"name": name, "age": age}



"""Utility Functions for Flask-Jeroboam.

Credits: this is essentially a fork of FastAPI's own utils.py
Original Source Code at https://github.com/tiangolo/fastapi
"""

import inspect
import re
import types
import typing
from collections.abc import Callable
from typing import Any, Union, get_args, get_origin


def _lenient_issubclass(cls: Any, class_or_tuple: Any) -> bool:
    """issubclass that returns False instead of raising TypeError."""
    try:
        return issubclass(cls, class_or_tuple)
    except TypeError:
        return False


def get_typed_signature(call: Callable[..., Any]) -> inspect.Signature:
    """Return a signature with string annotations resolved via get_type_hints."""
    signature = inspect.signature(call)
    globalns = getattr(call, "__globals__", {})
    try:
        hints = typing.get_type_hints(call, globalns=globalns, include_extras=True)
    except Exception:  # pragma: no cover
        hints = {}
    typed_params = [
        inspect.Parameter(
            name=param.name,
            kind=param.kind,
            default=param.default,
            annotation=hints.get(param.name, param.annotation),
        )
        for param in signature.parameters.values()
    ]
    return inspect.Signature(typed_params)


def get_typed_return_annotation(call: Callable[..., Any]) -> Any:
    """Return a typed return annotation."""
    try:
        hints = typing.get_type_hints(call, include_extras=True)
    except Exception:
        return None
    return hints.get("return", None)


def _unwrap_optional(annotation: Any) -> Any:
    """Return the inner type if annotation is Optional[X], else return it unchanged.

    Handles both typing.Optional[X] (Union) and the X | None syntax (types.UnionType
    in Python 3.10+).
    """
    origin = get_origin(annotation)
    is_union = origin is Union
    if not is_union and hasattr(types, "UnionType"):
        is_union = isinstance(annotation, types.UnionType)
    if is_union:
        non_none = [a for a in get_args(annotation) if a is not type(None)]
        if len(non_none) == 1:
            return non_none[0]
    return annotation


def is_sequence_field(field) -> bool:
    """Check if a field is a sequence field."""
    from typing import get_origin

    annotation = getattr(field, "annotation", None)
    if annotation is None:
        return False
    return get_origin(annotation) in (list, tuple, set, frozenset)


def _rename_query_params_keys(self, inbound_dict: dict, pattern: str) -> dict:
    """Rename keys in a dictionary."""
    frozen_inbound_dict = inbound_dict.copy()
    for old_key, value in frozen_inbound_dict.items():
        match = re.match(pattern, old_key)
        if match is not None:
            new_key = f"{match[1]}[]"
            new_value = {match[2]: value}
            new_array = inbound_dict.get(new_key, [])
            new_array.append(new_value)
            inbound_dict[new_key] = new_array
            del inbound_dict[old_key]
    return inbound_dict


def _throw_away_falthy_values(
    sparse_dict: dict[str, Any], keep: set | None = None
) -> dict[str, Any]:
    keep = keep or set()
    return {key: value for key, value in sparse_dict.items() if value or key in keep}


def _memoized_update_if_value(key: str, new_value: Any, orginal: dict[str, Any]):
    if new_value:
        orginal.setdefault(key, {}).update(new_value)


def _set_nested_defaults(
    original_dict: dict, keys: list[str], last_key: str, new_value: Any
) -> None:
    """Create a nested dictionary from a list of keys.

    _set_nested_defaults(
        original_dict=operation,
        keys=["responses", status_code, "content", route_response_media_type],
        last_key="schema",
        new_value={"$ref": "#/components/schemas/ResponseModel"},
    )
    is equivalent to:
    (operation.setdefault("responses", {})
     .setdefault(status_code, {})
     .setdefault("content", {})
     .setdefault(route_response_media_type, {})
    )["schema"] = {"$ref": "#/components/schemas/ResponseModel"}
    """
    for key in keys:
        original_dict.setdefault(key, {})
        original_dict = original_dict[key]
    original_dict[last_key] = new_value



import dataclasses
import traceback
from collections.abc import Callable
from functools import wraps
from typing import Any, TypeVar

from flask import Response
from flask.globals import current_app
from pydantic import BaseModel, RootModel, ValidationError
from typing_extensions import ParamSpec

from flask_jeroboam._constants import METHODS_DEFAULT_STATUS_CODE, NO_BODY_STATUS_CODES
from flask_jeroboam._utils import get_typed_return_annotation
from flask_jeroboam.exceptions import ResponseValidationError
from flask_jeroboam.responses import JSONResponse
from flask_jeroboam.typing import (
    HeadersValue,
    JeroboamBodyType,
    JeroboamResponseReturnValue,
    JeroboamRouteCallable,
    ResponseModel,
    Union,
)

# from flask_jeroboam.typing import TypedParams
# from flask_jeroboam.utils import get_typed_return_annotation


P = ParamSpec("P")
R = TypeVar("R")


class OutboundHandler:
    """The OutboundHandler handles outbound data of a request.

    More precisely, it prepare the content of the response, serializes it,
    determines the status code and form a proper Response object.

    The OutboundHandler will always be called if the view function
    returns without raising exception. Although if it receive an already
    well formed Response object, it will just return it.
    It's main purpose and public method is to add_outbound_handling_to
    """

    def __init__(
        self,
        view_func: Callable,
        configured_status_code: int | None,
        main_http_verb: str,
        options: dict[str, Any],
        response_class: type = JSONResponse,
    ):
        self.response_model = self._solve_response_model(view_func, options)
        self.configured_status_code = configured_status_code
        self.method_default_status_code = self._solve_default_status_code_by_http_verb(
            main_http_verb, configured_status_code
        )
        self.response_class = response_class
        self.response_description = options.pop(
            "response_description", "Successful Response"
        )

    @property
    def latent_status_code(self) -> int:
        """The status code that will be used if no status code is provided."""
        return self._solve_status_code(None)

    def add_outbound_handling_to(
        self, view_func: JeroboamRouteCallable
    ) -> JeroboamRouteCallable:
        """Add outbound handling to a view funcion."""

        @wraps(view_func)
        def outbound_handling(*args: Any, **kwargs: Any) -> JeroboamResponseReturnValue:
            """Coordinating the outbound data handling.

            It starts by executing the view_func with the given args and kwargs.
            At this point, the inbound handling has already parsed, validated
            and injected the incoming data in agrs and kwargs, provided it was
            configured to do so.
            If the initial response is already a well-formed Response object,
            it is returned as is.
            If not, it solve status code, serialize the content and finally
            build a Response object with it.
            It may raise a ValidationError if the outgoing data is not valid.

            Credits: this algorithm and subalgorithms are inspired by FastAPI.
            """
            initial_return_value = current_app.ensure_sync(view_func)(*args, **kwargs)
            if issubclass(initial_return_value.__class__, Response):
                return initial_return_value
            (
                returned_body,
                returned_status_code,
                headers,
            ) = self._unpack_view_function_return_value(initial_return_value)
            solved_status_code = self._solve_status_code(returned_status_code)
            if self._status_code_forbids_body(solved_status_code):
                return self._build_response(
                    status_code=solved_status_code, headers=headers
                )
            if self.response_model is None:
                return returned_body, solved_status_code, headers
            content = self._serialize_content(returned_body)
            return self._build_response(content, solved_status_code, headers=headers)

        return outbound_handling

    def _unpack_view_function_return_value(
        self, initial_return_value: JeroboamResponseReturnValue
    ) -> tuple[JeroboamBodyType, int | None, HeadersValue | None]:
        """Unpack the return value of the view function.

        Flask support various shapes of return values in their view function.
        We accomodate for theses shapes, but force them into a body/status_code
        /headers tuple.
        """
        if isinstance(initial_return_value, tuple):
            if len(initial_return_value) == 3:
                return initial_return_value  # type: ignore
            if len(initial_return_value) == 2:
                if isinstance(initial_return_value[1], int):
                    return (
                        initial_return_value[0],
                        initial_return_value[1],  # type:ignore
                        {},
                    )
                return (
                    initial_return_value[0],
                    None,
                    initial_return_value[1],  # type:ignore
                )
            raise TypeError(
                "The view function did not return a valid response tuple."
                " The tuple must have the form (body, status, headers),"
                " (body, status), or (body, headers)."
            )
        return initial_return_value, None, None  # type: ignore

    def _solve_response_model(
        self, view_func: Callable, options: dict[str, Any]
    ) -> ResponseModel | None:
        """Extract the Response Model from view function.

        The reponse_model must be a subclass of pydantic.BaseModel
        or a list of a of pydantic.BaseModel. Otherwise a type Error will be raised.
        It can be set in the options or a as return type annotation of the view
        function.
        The first one will take prescedence over the latter.
        Thus, to turn off registering a reponse_model, even when your
        view_function has a return annotation, set it to None.
        ..implementation details VS FastAPI:
            * Creating a model with root seems to do the trick. No need to generate
              a model_field from the response_model ?
            * Cloning the response_field doesn't seem to be necessary either
              to filter out unwanted data.
        """
        response_model: type | None = options.pop(
            "response_model", get_typed_return_annotation(view_func)
        )
        if response_model is None:
            return None

        if getattr(response_model, "__origin__", None) == list:
            field: type = response_model.__args__[0]  # type: ignore[attr-defined]

            class _AsList(RootModel[list[field]]):  # type: ignore[valid-type]
                pass

            _AsList.__name__ = f"{field.__name__}AsList"
            response_model = _AsList
        assert response_model is not None  # noqa: S101
        if not issubclass(response_model, BaseModel):
            raise TypeError(
                "The response model must be a subclass of pydantic.BaseModel."
            )
        return response_model

    def _solve_default_status_code_by_http_verb(
        self, http_verb: str, configured_status_code: int | None
    ) -> int | None:
        """Determines sensible default status code depending for the main HTTP VERB.

        TODO: Add a CONFIGURATION OPTION for this.
        """
        method_status_code = METHODS_DEFAULT_STATUS_CODE.get(http_verb, None)
        if method_status_code is None and configured_status_code is None:
            import warnings

            warnings.warn(
                f"No sensible default status code for verb {http_verb}. "
                "Maybe it's an exotic one. Make sure to set the status_code "
                "in the options and if you think we should add it, please file "
                "an issue.",
                UserWarning,
                stacklevel=2,
            )
        return method_status_code

    def _solve_status_code(self, returned_status_code: int | None) -> int:
        """Solve the status code from this Response.

        In order of priority:
        - the status code returned by the view function
        - the status code set in the options
        - the default status code for the main HTTP verb
        - the default status code for the ResponseClass


        TODO: Warning if conflict ?
        """
        candidates = [
            candidate
            for candidate in [
                returned_status_code,
                self.configured_status_code,
                self.method_default_status_code,
                self.response_class.default_status_code,  # type: ignore[attr-defined]
            ]
            if candidate is not None
        ]
        return candidates[0]

    def _serialize_content(self, content: JeroboamBodyType) -> Any:
        """Serialize the content of the response.

        # TODO: Algo de Sérialisation du Content de la Réponse.
        # Called only when we have a response_model
        # Sinon prepare_content qui renvoie un dict (option pour l'orm_mode)
        # On valide le dict avec le response_model et on collecte les erreurs
        # de validation
        # Option pour Couroutine or note
        # On raise un ValueError avec toutes les erreurs pas seulement la première.
        # on json_encode avec des options

        TODO: Ability to render_template ?
        TODO: Gestions des CoRoutine ?
        TODO: Plus d'options pour le passage en JSON ?
        TODO: Determine if we need by_alias, exclude_unset,
        exclude_defaults, exclude_none options ?
        """
        assert self.response_model is not None  # noqa: S101
        if issubclass(self.response_model, RootModel):
            content_to_validate = content
        else:
            content_to_validate = self._adapt_datastructure_of(content)  # type: ignore  # May raise ValueError
        try:
            validated = self.response_model.model_validate(content_to_validate)
        except ValidationError as error:
            raise ResponseValidationError(
                "A validation", error, traceback.format_exc()
            ) from error
        return validated.model_dump_json(by_alias=True)

    def _adapt_datastructure_of(
        self, content: JeroboamBodyType
    ) -> Union[dict, list[Any]]:
        """Prepare the content of the response.

        It basically accomodate for different datastructures and convert
        them to a dict.

        TODO: Do we support the ORM Mode ?$
        TODO: Better Name ?
        """
        if isinstance(content, dict):
            return content
        if isinstance(content, BaseModel):
            return content.model_dump()
        if isinstance(content, list):
            return [self._adapt_datastructure_of(item) for item in content]
        if dataclasses.is_dataclass(content) and not isinstance(content, type):
            return dataclasses.asdict(content)

        raise ValueError("Content must be a list, a dict, a dataclass, or a BaseModel.")

    def _build_response(
        self,
        content: str | None = None,
        status_code: int | None = None,
        headers: HeadersValue | None = None,
    ) -> Response:
        """Make a Response Object from content and status code, and passed_headers."""
        # Do we replace with a check on content is None ?
        if content is None:
            return self.response_class(status=status_code, headers=headers)
        return self.response_class(content, status=status_code, headers=headers)

    def _status_code_forbids_body(self, status_code: int) -> bool:
        """Check if the status code allows a body.

        Credits: Inspired by FastAPI.
        """
        return status_code < 200 or str(status_code) in NO_BODY_STATUS_CODES



"""Code Example 01 for README.md."""

from pydantic.fields import Field

from docs_src.readme._crud import get_wines
from flask_jeroboam import InboundModel, Jeroboam, OutboundModel

app = Jeroboam(__name__)


class GenericPagination(InboundModel):
    page: int = Field(1, ge=1)
    per_page: int = Field(10, ge=1, le=100)

    @property
    def offset(self) -> int:
        return (self.page - 1) * self.per_page


class WineOut(OutboundModel):
    cuvee: str
    appellation: str


@app.get("/ping")
def ping():
    return "pong"


@app.get("/wines", response_model=list[WineOut])
def read_wine_list(pagination: GenericPagination, search: str | None):
    return get_wines(pagination, search)


if __name__ == "__main__":
    app.run(port=5000)



from enum import Enum


class MethodEnum(str, Enum):
    """List of HTTP Methods."""

    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"


REF_PREFIX = "#/components/schemas/"
METHODS_WITH_BODY = {"POST", "PUT", "PATCH", "DELETE"}
NO_BODY_STATUS_CODES = {"204", "205", "304"}

VALIDATION_ERROR_DEFINITION = {
    "title": "ValidationError",
    "type": "object",
    "properties": {
        "loc": {
            "title": "Location",
            "type": "array",
            "items": {"anyOf": [{"type": "string"}, {"type": "integer"}]},
        },
        "msg": {"title": "Message", "type": "string"},
        "type": {"title": "Error Type", "type": "string"},
    },
    "required": ["loc", "msg", "type"],
}

VALIDATION_ERROR_RESPONSE_DEFINITION = {
    "title": "HTTPValidationError",
    "type": "object",
    "properties": {
        "detail": {
            "title": "Detail",
            "type": "array",
            "items": {"$ref": f"{REF_PREFIX}ValidationError"},
        }
    },
}

status_code_ranges: dict[str, str] = {
    "1XX": "Information",
    "2XX": "Success",
    "3XX": "Redirection",
    "4XX": "Client Error",
    "5XX": "Server Error",
    "DEFAULT": "Default Response",
}


METHODS_DEFAULT_STATUS_CODE = {
    "GET": 200,
    "HEAD": 200,
    "POST": 201,
    "PUT": 201,
    "DELETE": 204,
    "CONNECT": 200,
    "OPTIONS": 200,
    "TRACE": 200,
    "PATCH": 200,
}

NO_BODY_STATUS_CODES = {"204", "205", "304"}



"""Utlility functions for creating the OpenAPI doc.

Credits: This is a Fork of FastAPI's openapi/utils.py
"""

import warnings
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    cast,
)

from pydantic import BaseModel

from flask_jeroboam._constants import (
    NO_BODY_STATUS_CODES,
    REF_PREFIX,
    VALIDATION_ERROR_DEFINITION,
    VALIDATION_ERROR_RESPONSE_DEFINITION,
)
from flask_jeroboam._utils import (
    _lenient_issubclass,
    _set_nested_defaults,
    _throw_away_falthy_values,
    _unwrap_optional,
)

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Sequence

    from flask_jeroboam.rule import JeroboamRule
    from flask_jeroboam.view import JeroboamView
    from flask_jeroboam.view_arguments.arguments import BodyArgument, ParameterArgument
    from flask_jeroboam.view_arguments.solved import SolvedArgument


def _get_param_schema(param: "SolvedArgument") -> dict[str, Any]:
    from pydantic import TypeAdapter

    # For Optional[X] (union with None), use just the inner type for schema generation
    # so that we get {"type": "integer"} rather than {"anyOf": [...]}
    inner_annotation = _unwrap_optional(param.annotation)
    if inner_annotation is not param.annotation:
        schema = TypeAdapter(inner_annotation).json_schema(
            ref_template=REF_PREFIX + "{model}"
        )
    else:
        schema = param._type_adapter.json_schema(ref_template=REF_PREFIX + "{model}")
    # Remove "default": None from schema — None defaults are implicit in optional params
    if schema.get("default") is None:
        schema.pop("default", None)
    schema.setdefault("title", param.alias.replace("_", " ").title())
    return schema


def _get_openapi_operation_parameters(
    *,
    all_route_params: "Sequence[SolvedArgument]",
) -> list[dict[str, Any]]:
    parameters = []
    for param in all_route_params:
        field_info = param.field_info
        field_info = cast("ParameterArgument", field_info)
        if getattr(param, "include_in_schema", True) is False:
            continue
        parameter = _throw_away_falthy_values(
            {
                "name": param.alias,
                "in": field_info.location.value,
                "required": param.required,
                "schema": _get_param_schema(param),
                "deprecated": field_info.deprecated,
                "description": field_info.description,
            },
            keep={"name", "in", "required"},
        )
        parameters.append(parameter)
    return parameters


def _get_openapi_operation_metadata(
    *, rule: "JeroboamRule", method: str, operation_ids: set[str]
) -> dict[str, Any]:
    operation_id = rule.operation_id or rule.unique_id or rule.endpoint
    if operation_id in operation_ids:
        message = (
            f"Duplicate Operation ID {operation_id} for function " + f"{rule.endpoint}"
        )
        warnings.warn(message, UserWarning, stacklevel=2)
    operation_ids.add(operation_id)
    operation: dict[str, Any] = _throw_away_falthy_values(
        {
            "tags": rule.tags,
            "description": rule.description,
            "summary": rule.summary,
            "deprecated": rule.deprecated,
            "operationId": operation_id,
        }
    )
    return operation


def _get_body_schema(body_field: "SolvedArgument") -> dict[str, Any]:
    """Select the correct JSON schema representation for a request body annotation."""
    annotation = body_field.annotation
    is_synthetic = _lenient_issubclass(
        annotation, BaseModel
    ) and annotation.__name__.endswith("request_body_as_model")
    if is_synthetic:
        body_schema = cast("type[BaseModel]", annotation).model_json_schema(
            ref_template=REF_PREFIX + "{model}"
        )
        body_schema.pop("$defs", None)
        return body_schema
    if _lenient_issubclass(annotation, BaseModel):
        return {"$ref": f"{REF_PREFIX}{annotation.__name__}"}
    return body_field._type_adapter.json_schema(ref_template=REF_PREFIX + "{model}")


def _get_openapi_operation_request_body(
    *,
    body_field: "SolvedArgument | None",
) -> dict[str, Any] | None:
    if body_field is None:  # pragma: no cover
        return None
    field_info = cast("BodyArgument", body_field.field_info)
    body_schema = _get_body_schema(body_field)
    request_body_oai: dict[str, Any] = {}
    if body_field.required:
        request_body_oai["required"] = True
    request_body_oai["content"] = {field_info.media_type: {"schema": body_schema}}
    return request_body_oai


def _get_response_schema(
    response_model: type | None,
) -> dict[str, Any]:
    if response_model is None:
        return {}
    if _lenient_issubclass(response_model, BaseModel):
        return {"$ref": f"{REF_PREFIX}{response_model.__name__}"}
    return {}  # pragma: no cover


def _add_responses(
    operation,
    jeroboam_view,
):
    definitions: dict[str, Any] = {}
    status_code = str(jeroboam_view.outbound_handler.latent_status_code)
    current_response_class = jeroboam_view.outbound_handler.response_class
    route_response_media_type: str | None = current_response_class.default_mimetype

    _set_nested_defaults(
        original_dict=operation,
        keys=["responses", status_code],
        last_key="description",
        new_value=jeroboam_view.outbound_handler.response_description,
    )

    if route_response_media_type and status_code not in NO_BODY_STATUS_CODES:
        _set_nested_defaults(
            original_dict=operation,
            keys=["responses", status_code, "content", route_response_media_type],
            last_key="schema",
            new_value=_get_response_schema(
                jeroboam_view.outbound_handler.response_model,
            ),
        )

    # TODO: possibilité de configurer le status code
    if jeroboam_view.inbound_handler.is_valid and all(
        status not in operation["responses"] for status in {"400", "4XX", "default"}
    ):
        operation["responses"]["400"] = {
            "description": "Validation Error",
            "content": {
                "application/json": {
                    "schema": {"$ref": f"{REF_PREFIX}HTTPValidationError"}
                }
            },
        }
        definitions.update(
            {
                "ValidationError": VALIDATION_ERROR_DEFINITION,
                "HTTPValidationError": VALIDATION_ERROR_RESPONSE_DEFINITION,
            }
        )
    return operation, definitions


def _build_openapi_path_item(
    *,
    rule: "JeroboamRule",
    jeroboam_view: "JeroboamView",
    operation_ids: set[str],
) -> tuple[dict[str, Any], dict[str, Any], dict[str, Any]]:
    """Build OpenApi path from our rule Object.

    TODO: Gérer les Callbacks ?
    """
    operation = _get_openapi_operation_metadata(
        rule=rule, method=jeroboam_view.main_method, operation_ids=operation_ids
    )

    # On gère les paramètres
    if parameters := _get_openapi_operation_parameters(
        all_route_params=jeroboam_view.parameters,
    ):
        all_parameters = {(param["in"], param["name"]): param for param in parameters}
        required_parameters = {
            (param["in"], param["name"]): param
            for param in parameters
            if param.get("required")
        }
        # Make sure required definitions of the same parameter take precedence
        # over non-required definitions
        all_parameters.update(required_parameters)
        operation["parameters"] = list(all_parameters.values())
    # On gère le request Body
    if jeroboam_view.has_request_body:
        operation["requestBody"] = _get_openapi_operation_request_body(
            body_field=jeroboam_view.inbound_handler.body_field(rule.unique_id),
        )

    operation, definitions = _add_responses(
        operation,
        jeroboam_view,
    )
    operation.update(rule.openapi_extra or {})
    return {jeroboam_view.main_method: operation}, {}, definitions


def _get_model_definitions(
    *,
    flat_models: set,
) -> dict[str, Any]:
    definitions: dict[str, Any] = {}
    for model in flat_models:
        m_schema = model.model_json_schema(ref_template=REF_PREFIX + "{model}")
        nested = m_schema.pop("$defs", {})
        definitions.update(nested)
        if "description" in m_schema:
            m_schema["description"] = m_schema["description"].split("\f")[0]
        definitions[model.__name__] = m_schema
    return definitions


def _collect_models_from_view(
    jeroboam_view: "JeroboamView", rule: "JeroboamRule", models: set
) -> None:
    """Add response and body models from a single view to the models set."""
    response_model = jeroboam_view.outbound_handler.response_model
    if response_model is not None:
        models.add(response_model)
    body_field = jeroboam_view.inbound_handler.body_field(rule.unique_id)
    if (
        body_field is not None
        and _lenient_issubclass(body_field.annotation, BaseModel)
        and not body_field.annotation.__name__.endswith("request_body_as_model")
    ):
        models.add(body_field.annotation)


def _get_flat_models_from_jeroboam_views(
    jeroboam_views: list[Optional["JeroboamView"]], rules: list["JeroboamRule"]
):
    models: set = set()
    for jeroboam_view, rule in zip(jeroboam_views, rules):
        if jeroboam_view is None:
            continue
        if not getattr(jeroboam_view, "include_in_openapi", True):
            continue
        _collect_models_from_view(jeroboam_view, rule, models)
    return models



"""A Test Blueprint for testing Query Params.

The corresponding test can be found in tests/test_inbound/test_query
"""

from flask_jeroboam import Blueprint, Query
from tests.app_test.models.inbound import (
    ModelWithListIn,
    OptionalModelIn,
    QueryStringWithList,
    SimpleModelIn,
)
from tests.app_test.models.outbound import ModelWithListOut

router = Blueprint("query_params_router", __name__, tags=["Query"])


@router.get("/query/frozenset")
def get_query_type_frozenset(query: frozenset[int] = Query(...)):
    return {"query": ",".join(map(str, sorted(query)))}


@router.get("/query")
def get_query(query):
    return {"query": query}


@router.get("/query/optional")
def get_query_optional(query=None):
    return {"query": query}


@router.get("/query/int")
def get_query_type(query: int):
    return {"query": query}


@router.get("/query/int/optional")
def get_query_type_optional(query: int | None = None):
    return {"query": query}


@router.get("/query/int/default")
def get_query_type_int_default(query: int = 10):
    return {"query": query}


@router.get("/query/param")
def get_query_param(query=Query(default=None)):
    return {"query": query}


@router.get("/query/param-required")
def get_query_param_required(query=Query()):
    return {"query": query}


@router.get("/query/param-required/int")
def get_query_param_required_type(query: int = Query()):
    return {"query": query}


@router.get("/enum-status-code", status_code=201)
def get_enum_status_code():
    return {}


@router.get("/query/base_model")
def get_base_model(payload: SimpleModelIn):
    """Base Model as Query Param."""
    return payload.model_dump_json()


@router.get("/query/base_model/forward_ref")
def get_base_model_as_forward_ref(payload: "SimpleModelIn"):
    return payload.model_dump_json()


@router.get("/query/list_of_strings")
def get_list_of_strings(query_string: QueryStringWithList):
    return query_string.model_dump_json()


@router.get("/query/optional_model")
def get_optional_param(payload: OptionalModelIn | None):
    return payload.model_dump_json() if payload else {}


@router.get("/query/special_pattern", response_model=ModelWithListOut)
def read_items(payload: ModelWithListIn):
    return payload.model_dump_json()



"""A Test Blueprint for testing Path Params.

The corresponding test can be found in tests/test_inbound/test_path
"""

from flask_jeroboam import Blueprint, Path

router = Blueprint("path_params_router", __name__, tags=["Path"])


@router.get("/path/<item_id>")
def get_id(item_id):
    return {"item_id": item_id}


@router.get("/path/str/<item_id>")
def get_str_id(item_id: str):
    return {"item_id": item_id}


@router.get("/path/int/<item_id>")
def get_int_id(item_id: int):
    return {"item_id": item_id}


@router.get("/path/float/<item_id>")
def get_float_id(item_id: float):
    return {"item_id": item_id}


@router.get("/path/bool/<item_id>")
def get_bool_id(item_id: bool):
    return {"item_id": item_id}


@router.get("/path/param/<item_id>")
def get_path_param_id(item_id: str = Path()):
    return {"item_id": item_id}


@router.get("/path/param-required/<item_id>")
def get_path_param_required_id(item_id: str = Path()):
    return {"item_id": item_id}


@router.get("/path/param-minlength/<item_id>")
def get_path_param_min_length(item_id: str = Path(min_length=3)):
    return {"item_id": item_id}


@router.get("/path/param-maxlength/<item_id>")
def get_path_param_max_length(item_id: str = Path(max_length=3)):
    return {"item_id": item_id}


@router.get("/path/param-min_maxlength/<item_id>")
def get_path_param_min_max_length(item_id: str = Path(max_length=3, min_length=2)):
    return {"item_id": item_id}


@router.get("/path/param-gt/<item_id>")
def get_path_param_gt(item_id: float = Path(gt=3)):
    return {"item_id": item_id}


@router.get("/path/param-gt0/<item_id>")
def get_path_param_gt0(item_id: float = Path(gt=0)):
    return {"item_id": item_id}


@router.get("/path/param-ge/<item_id>")
def get_path_param_ge(item_id: float = Path(ge=3)):
    return {"item_id": item_id}


@router.get("/path/param-lt/<item_id>")
def get_path_param_lt(item_id: float = Path(lt=3)):
    return {"item_id": item_id}


@router.get("/path/param-lt0/<item_id>")
def get_path_param_lt0(item_id: float = Path(lt=0)):
    return {"item_id": item_id}


@router.get("/path/param-le/<item_id>")
def get_path_param_le(item_id: float = Path(le=3)):
    return {"item_id": item_id}


@router.get("/path/param-lt-gt/<item_id>")
def get_path_param_lt_gt(item_id: float = Path(lt=3, gt=1)):
    return {"item_id": item_id}


@router.get("/path/param-le-ge/<item_id>")
def get_path_param_le_ge(item_id: float = Path(le=3, ge=1)):
    return {"item_id": item_id}


@router.get("/path/param-lt-int/<item_id>")
def get_path_param_lt_int(item_id: int = Path(lt=3)):
    return {"item_id": item_id}


@router.get("/path/param-gt-int/<item_id>")
def get_path_param_gt_int(item_id: int = Path(gt=3)):
    return {"item_id": item_id}


@router.get("/path/param-le-int/<item_id>")
def get_path_param_le_int(item_id: int = Path(le=3)):
    return {"item_id": item_id}


@router.get("/path/param-ge-int/<item_id>")
def get_path_param_ge_int(item_id: int = Path(ge=3)):
    return {"item_id": item_id}


@router.get("/path/param-lt-gt-int/<item_id>")
def get_path_param_lt_gt_int(item_id: int = Path(lt=3, gt=1)):
    return {"item_id": item_id}


@router.get("/path/param-le-ge-int/<item_id>")
def get_path_param_le_ge_int(item_id: int = Path(le=3, ge=1)):
    return {"item_id": item_id}


@router.get("/path/with_converter/<item_id>")
def get_with_preproc_id(item_id):
    return {"item_id": item_id}


@router.get("/path/with_converter/str/<string:item_id>")
def get_with_preproc_str_id(item_id: str):
    return {"item_id": item_id}


@router.get("/path/with_converter/int/<int:item_id>")
def get_with_preproc_int_id(item_id: int):
    return {"item_id": item_id}


@router.get("/path/with_converter/float/<float:item_id>")
def get_with_preproc_float_id(item_id: float):
    return {"item_id": item_id}


@router.get("/path/with_converter/param/<string:item_id>")
def get_with_preproc_path_param_id(item_id: str = Path()):
    return {"item_id": item_id}


@router.get("/path/with_converter/param-required/<string:item_id>")
def get_with_preproc_path_param_required_id(item_id: str = Path()):
    return {"item_id": item_id}


@router.get("/path/with_converter/param-minlength/<string:item_id>")
def get_with_preproc_path_param_min_length(item_id: str = Path(min_length=3)):
    return {"item_id": item_id}


@router.get("/path/with_converter/param-maxlength/<string:item_id>")
def get_with_preproc_path_param_max_length(item_id: str = Path(max_length=3)):
    return {"item_id": item_id}


@router.get("/path/with_converter/param-min_maxlength/<string:item_id>")
def get_with_preproc_path_param_min_max_length(
    item_id: str = Path(max_length=3, min_length=2),
):
    return {"item_id": item_id}


@router.get("/path/with_converter/param-gt/<float:item_id>")
def get_with_preproc_path_param_gt(item_id: float = Path(gt=3)):
    return {"item_id": item_id}


@router.get("/path/with_converter/param-gt0/<float:item_id>")
def get_with_preproc_path_param_gt0(item_id: float = Path(gt=0)):
    return {"item_id": item_id}


@router.get("/path/with_converter/param-ge/<float:item_id>")
def get_with_preproc_path_param_ge(item_id: float = Path(ge=3)):
    return {"item_id": item_id}


@router.get("/path/with_converter/param-lt/<float:item_id>")
def get_with_preproc_path_param_lt(item_id: float = Path(lt=3)):
    return {"item_id": item_id}


@router.get("/path/with_converter/param-lt0/<float(signed=True):item_id>")
def get_with_preproc_path_param_lt0(item_id: float = Path(lt=0)):
    return {"item_id": item_id}


@router.get("/path/with_converter/param-le/<float:item_id>")
def get_with_preproc_path_param_le(item_id: float = Path(le=3)):
    return {"item_id": item_id}


@router.get("/path/with_converter/param-lt-gt/<float:item_id>")
def get_with_preproc_path_param_lt_gt(item_id: float = Path(lt=3, gt=1)):
    return {"item_id": item_id}


@router.get("/path/with_converter/param-le-ge/<float:item_id>")
def get_with_preproc_path_param_le_ge(item_id: float = Path(le=3, ge=1)):
    return {"item_id": item_id}


@router.get("/path/with_converter/param-lt-int/<int:item_id>")
def get_with_preproc_path_param_lt_int(item_id: int = Path(lt=3)):
    return {"item_id": item_id}


@router.get("/path/with_converter/param-gt-int/<int:item_id>")
def get_with_preproc_path_param_gt_int(item_id: int = Path(gt=3)):
    return {"item_id": item_id}


@router.get("/path/with_converter/param-le-int/<int:item_id>")
def get_with_preproc_path_param_le_int(item_id: int = Path(le=3)):
    return {"item_id": item_id}


@router.get("/path/with_converter/param-ge-int/<int:item_id>")
def get_with_preproc_path_param_ge_int(item_id: int = Path(ge=3)):
    return {"item_id": item_id}


@router.get("/path/with_converter/param-lt-gt-int/<int:item_id>")
def get_with_preproc_path_param_lt_gt_int(item_id: int = Path(lt=3, gt=1)):
    return {"item_id": item_id}


@router.get("/path/with_converter/param-le-ge-int/<int:item_id>")
def get_with_preproc_path_param_le_ge_int(item_id: int = Path(le=3, ge=1)):
    return {"item_id": item_id}



"""OpenAPI Specification models.

Credits: A fork of FASTAPI's openapi/models.py
"""

from enum import Enum
from typing import Any, Optional, Union

from pydantic import AnyUrl, BaseModel, ConfigDict, Field

from flask_jeroboam._logger import logger

try:  # pragma: no cover
    import email_validator  # type: ignore
    from pydantic import EmailStr  # type: ignore

    assert email_validator  # noqa: S101
except ImportError:  # pragma: no cover
    from pydantic import GetCoreSchemaHandler
    from pydantic_core import core_schema as _core_schema

    class EmailStr(str):  # type: ignore
        """An empty EmailStr to warn users if email-validator is not installed."""

        @classmethod
        def __get_pydantic_core_schema__(
            cls, source_type: Any, handler: GetCoreSchemaHandler
        ) -> Any:
            return _core_schema.no_info_plain_validator_function(cls.validate)

        @classmethod
        def validate(cls, v: Any) -> str:
            """Raises a warning if email-validator is not installed."""
            logger.warning(
                "email-validator not installed, email fields will be treated as str.\n"
                "To install, run: pip install pydantic[email]"
            )
            return str(v)


class Contact(BaseModel):
    name: str | None = None
    url: AnyUrl | None = None
    email: EmailStr | None = None

    model_config = ConfigDict(extra="allow")


class License(BaseModel):
    name: str
    url: AnyUrl | None = None

    model_config = ConfigDict(extra="allow")


class Info(BaseModel):
    title: str
    description: str | None = None
    terms_of_service: str | None = Field(None, alias="termsOfService")
    contact: Contact | None = None
    license: License | None = None
    version: str

    model_config = ConfigDict(extra="allow")


class ServerVariable(BaseModel):
    enum: list[str] | None = None
    default: str
    description: str | None = None

    model_config = ConfigDict(extra="allow")


class Server(BaseModel):
    url: AnyUrl | str
    description: str | None = None
    variables: dict[str, ServerVariable] | None = None

    model_config = ConfigDict(extra="allow")


class Reference(BaseModel):
    ref: str = Field(alias="$ref")


class Discriminator(BaseModel):
    property_name: str
    mapping: dict[str, str] | None = None


class XML(BaseModel):
    name: str | None = None
    namespace: str | None = None
    prefix: str | None = None
    attribute: bool | None = None
    wrapped: bool | None = None

    model_config = ConfigDict(extra="allow")


class ExternalDocumentation(BaseModel):
    description: str | None = None
    url: AnyUrl

    model_config = ConfigDict(extra="allow")


class Schema(BaseModel):
    ref: str | None = Field(default=None, alias="$ref")
    title: str | None = None
    multiple_of: float | None = Field(None, alias="multipleOf")
    maximum: float | None = None
    exclusive_maximum: float | None = Field(None, alias="exclusiveMaximum")
    minimum: float | None = None
    exclusive_minimum: float | None = Field(None, alias="exclusiveMinimum")
    max_length: int | None = Field(default=None, ge=0, alias="maxLength")
    min_length: int | None = Field(default=None, ge=0, alias="minLength")
    pattern: str | None = None
    max_items: int | None = Field(default=None, ge=0, alias="maxItems")
    min_items: int | None = Field(default=None, ge=0, alias="minItems")
    unique_items: bool | None = Field(None, alias="uniqueItems")
    max_properties: int | None = Field(default=None, ge=0, alias="maxProperties")
    min_properties: int | None = Field(default=None, ge=0, alias="minProperties")
    required: list[str] | None = None
    enum: list[Any] | None = None
    type: str | None = None
    all_of: list["Schema"] | None = Field(None, alias="allOf")
    one_of: list["Schema"] | None = Field(None, alias="oneOf")
    any_of: list["Schema"] | None = Field(None, alias="anyOf")
    not_: Optional["Schema"] = Field(default=None, alias="not")
    items: Union["Schema", list["Schema"]] | None = None
    properties: dict[str, "Schema"] | None = None
    additional_properties: Union["Schema", Reference, bool] | None = Field(
        None, alias="additionalProperties"
    )
    description: str | None = None
    format: str | None = None
    default: Any | None = None
    nullable: bool | None = None
    discriminator: Discriminator | None = None
    read_only: bool | None = Field(None, alias="readOnly")
    write_only: bool | None = Field(None, alias="writeOnly")
    xml: XML | None = None
    external_docs: ExternalDocumentation | None = Field(None, alias="externalDocs")
    example: Any | None = None
    deprecated: bool | None = None

    model_config = ConfigDict(extra="allow")


class Example(BaseModel):
    summary: str | None = None
    description: str | None = None
    value: Any | None = None
    external_value: AnyUrl | None = Field(None, alias="externalValue")

    model_config = ConfigDict(extra="allow")


class ParameterInType(Enum):
    query = "query"
    header = "header"
    path = "path"
    cookie = "cookie"


class Encoding(BaseModel):
    content_type: str | None = Field(None, alias="contentType")
    headers: dict[str, Union["Header", Reference]] | None = None
    style: str | None = None
    explode: bool | None = None
    allow_reserved: bool | None = Field(None, alias="allowReserved")

    model_config = ConfigDict(extra="allow")


class MediaType(BaseModel):
    schema_: Schema | Reference | None = Field(default=None, alias="schema")
    example: Any | None = None
    examples: dict[str, Example | Reference] | None = None
    encoding: dict[str, Encoding] | None = None

    model_config = ConfigDict(extra="allow")


class ParameterBase(BaseModel):
    description: str | None = None
    required: bool | None = None
    deprecated: bool | None = None
    # Serialization rules for simple scenarios
    style: str | None = None
    explode: bool | None = None
    allow_reserved: bool | None = Field(None, alias="allowReserved")
    schema_: Schema | Reference | None = Field(default=None, alias="schema")
    example: Any | None = None
    examples: dict[str, Example | Reference] | None = None
    # Serialization rules for more complex scenarios
    content: dict[str, MediaType] | None = None

    model_config = ConfigDict(extra="allow")


class Parameter(ParameterBase):
    name: str
    in_: ParameterInType = Field(alias="in")


class Header(ParameterBase):
    pass


class RequestBody(BaseModel):
    description: str | None = None
    content: dict[str, MediaType]
    required: bool | None = None

    model_config = ConfigDict(extra="allow")


class Link(BaseModel):
    operation_ref: str | None = Field(None, alias="operationRef")
    operation_id: str | None = Field(None, alias="operationId")
    parameters: dict[str, Any | str] | None = None
    request_body: Any | str | None = Field(None, alias="requestBody")
    description: str | None = None
    server: Server | None = None

    model_config = ConfigDict(extra="allow")


class Response(BaseModel):
    description: str
    headers: dict[str, Header | Reference] | None = None
    content: dict[str, MediaType] | None = None
    links: dict[str, Link | Reference] | None = None

    model_config = ConfigDict(extra="allow")


class Operation(BaseModel):
    tags: list[str] | None = None
    summary: str | None = None
    description: str | None = None
    external_docs: ExternalDocumentation | None = Field(None, alias="externalDocs")
    operation_id: str | None = Field(None, alias="operationId")
    parameters: list[Parameter | Reference] | None = None
    request_body: RequestBody | Reference | None = Field(None, alias="requestBody")
    # Using Any for Specification Extensions
    responses: dict[str, Response | Any]
    callbacks: dict[str, dict[str, "PathItem"] | Reference] | None = None
    deprecated: bool | None = None
    security: list[dict[str, list[str]]] | None = None
    servers: list[Server] | None = None

    model_config = ConfigDict(extra="allow")


class PathItem(BaseModel):
    ref: str | None = Field(default=None, alias="$ref")
    summary: str | None = None
    description: str | None = None
    get: Operation | None = None
    put: Operation | None = None
    post: Operation | None = None
    delete: Operation | None = None
    options: Operation | None = None
    head: Operation | None = None
    patch: Operation | None = None
    trace: Operation | None = None
    servers: list[Server] | None = None
    parameters: list[Parameter | Reference] | None = None

    model_config = ConfigDict(extra="allow")


class SecuritySchemeType(Enum):
    api_key = "apiKey"
    http = "http"
    oauth2 = "oauth2"
    open_id_connect = "openIdConnect"


class SecurityBase(BaseModel):
    type_: SecuritySchemeType = Field(alias="type")
    description: str | None = None

    model_config = ConfigDict(extra="allow")


class APIKeyIn(Enum):
    query = "query"
    header = "header"
    cookie = "cookie"


class APIKey(SecurityBase):
    type_: SecuritySchemeType = Field(SecuritySchemeType.api_key, alias="type")
    in_: APIKeyIn = Field(alias="in")
    name: str


class HTTPBase(SecurityBase):
    type_: SecuritySchemeType = Field(SecuritySchemeType.http, alias="type")
    scheme: str


class HTTPBearer(HTTPBase):
    scheme: str = "bearer"
    bearer_format: str | None = Field(None, alias="bearerFormat")


class OAuthFlow(BaseModel):
    refresh_url: str | None = Field(None, alias="refreshUrl")
    scopes: dict[str, str] = {}

    model_config = ConfigDict(extra="allow")


class OAuthFlowImplicit(OAuthFlow):
    authorization_url: str = Field(..., alias="authorizationUrl")


class OAuthFlowPassword(OAuthFlow):
    token_url: str = Field(..., alias="tokenUrl")


class OAuthFlowClientCredentials(OAuthFlow):
    token_url: str = Field(..., alias="tokenUrl")


class OAuthFlowAuthorizationCode(OAuthFlow):
    authorization_url: str = Field(..., alias="authorizationUrl")
    token_url: str = Field(..., alias="tokenUrl")


class OAuthFlows(BaseModel):
    implicit: OAuthFlowImplicit | None = None
    password: OAuthFlowPassword | None = None
    client_credentials: OAuthFlowClientCredentials | None = Field(
        None, alias="clientCredentials"
    )
    authorization_code: OAuthFlowAuthorizationCode | None = Field(
        None, alias="authorizationCode"
    )

    model_config = ConfigDict(extra="allow")


class OAuth2(SecurityBase):
    type_: SecuritySchemeType = Field(SecuritySchemeType.oauth2, alias="type")
    flows: OAuthFlows


class OpenIdConnect(SecurityBase):
    type_: SecuritySchemeType = Field(SecuritySchemeType.open_id_connect, alias="type")
    open_id_connect_url: str = Field(..., alias="openIdConnectUrl")


SecurityScheme = APIKey | HTTPBase | OAuth2 | OpenIdConnect | HTTPBearer


class Components(BaseModel):
    schemas: dict[str, Schema | Reference] | None = None
    responses: dict[str, Response | Reference] | None = None
    parameters: dict[str, Parameter | Reference] | None = None
    examples: dict[str, Example | Reference] | None = None
    request_bodies: dict[str, RequestBody | Reference] | None = Field(
        default=None, alias="requestBodies"
    )
    headers: dict[str, Header | Reference] | None = None
    security_schemes: dict[str, SecurityScheme | Reference] | None = Field(
        default=None, alias="securitySchemes"
    )
    links: dict[str, Link | Reference] | None = None
    # Using Any for Specification Extensions
    callbacks: dict[str, dict[str, PathItem] | Reference | Any] | None = None

    model_config = ConfigDict(extra="allow")


class Tag(BaseModel):
    name: str
    description: str | None = None
    external_docs: ExternalDocumentation | None = Field(None, alias="externalDocs")

    model_config = ConfigDict(extra="allow")


class OpenAPI(BaseModel):
    openapi: str
    info: Info
    servers: list[Server] | None = None
    # Using Any for Specification Extensions
    paths: dict[str, PathItem | Any]
    components: Components | None = None
    security: list[dict[str, list[str]]] | None = None
    tags: list[Tag] | None = None
    external_docs: ExternalDocumentation | None = Field(
        default=None, alias="externalDocs"
    )

    model_config = ConfigDict(extra="allow")


Schema.model_rebuild()
Operation.model_rebuild()
Encoding.model_rebuild()



"""Code Sample for Inbound Features Documentation."""

from pydantic import BaseModel

from flask_jeroboam import Body, Cookie, Jeroboam, Path, Query

app = Jeroboam("Jeroboam Inbound Features App")
app.init_app()


class Item(BaseModel):
    name: str
    count: int


@app.get("/implicit_location_is_query_string")
def get_implicit_argument(page: int):
    return f"Received Page Argument is : {page}"


@app.get("/explicit_location_is_query_string")
def get_explicit_argument(page: int = Query()):
    return f"Received Page Argument is : {page}"


@app.get("/explicit_location_is_query_string_and_cookie")
def get_explicit_arguments(page: int = Query(), username: str = Cookie()):
    return f"Received Page Argument is : {page}. Username is : {username}"


@app.get("/implicit_and_explicit")
def get_implicit_and_explicit_arguments(page: int, username: str = Cookie()):
    return f"Received Page Argument is : {page}. Username is : {username}"


@app.post("/implicit_location_is_body")
def post_implicit_argument(page: int):
    return f"Received Page Argument is : {page}"


@app.post("/explicit_location_is_body")
def post_explicit_argument(page: int = Body()):
    return f"Received Page Argument is : {page}"


@app.get("/item/<int:id>/implicit")
def get_path_parameter(id: int):
    return f"Received id Argument is : {id}"


@app.post("/item/<int:id>/implicit")
def post_path_parameter(id: int):
    return f"Received id Argument is : {id}"


@app.post("/item/<int:id>/explicit")
def get_explicit_path_parameter(id: int = Path()):
    return f"Received id Argument is : {id}"


@app.get("/implicit_location_with_default_value")
def get_implicit_argument_with_default(page: int = 1):
    return f"Received Page Argument is : {page}"


@app.get("/explicit_location_with_default_value")
def get_explicit_argument_with_default(page: int = Query(1)):
    return f"Received Page Argument is : {page}"


@app.get("/defining_type_with_type_hints")
def get_type_hints(page: int, search: list[str], item: Item, price=Query(float)):
    return (
        f"Received arguments are :\n"
        f"page : {page}\n"
        f"search : {search}\n"
        f"price : {price}\n"
        f"item : {item}"
    )


@app.get("/argument_with_validation")
def get_validation_option(page: int = Query(ge=1)):
    return f"Received page argument is : {page}"


if __name__ == "__main__":
    app.run(host="localhost", port=5000)



import inspect
import re
import typing as t
from collections.abc import Callable
from functools import wraps
from typing import Any

from pydantic import BaseModel, create_model
from pydantic_core import PydanticUndefined
from typing_extensions import ParamSpec

from flask_jeroboam._utils import _lenient_issubclass, get_typed_signature
from flask_jeroboam.exceptions import InvalidRequest
from flask_jeroboam.typing import JeroboamResponseReturnValue, JeroboamRouteCallable
from flask_jeroboam.view_arguments.arguments import (
    ArgumentLocation,
    BodyArgument,
    ViewArgument,
    get_argument_class,
)
from flask_jeroboam.view_arguments.functions import Body, File, Form
from flask_jeroboam.view_arguments.solved import SolvedArgument

F = t.TypeVar("F", bound=t.Callable[..., t.Any])
R = t.TypeVar("R", bound=t.Any)
P = ParamSpec("P")
T = t.TypeVar("T", bound=t.Any)


pattern = r"(.*)\[(.+)\]$"


def _unpack_body_values(body_value: dict) -> dict:
    """Unpack a synthetic body model's fields back into individual view kwargs."""
    values: dict = {}
    for model in body_value.values():
        if isinstance(model, BaseModel):  # pragma: no branch
            for field_name in type(model).model_fields:
                values[field_name] = getattr(model, field_name)
    return values


class InboundHandler:
    """The InboundHandler handles inbound data of a request.

    More precisely, it parses the incoming data, validates it, and injects it into the
    view function. It is also responsible for raising an InvalidRequest exception.
    The InboundHandler will only be called if the view function has type-annotated
    parameters.



    #TODO: Get Better at laying Out Levels of the Algorythm. Most Likely in the View
    # class.
    # And Moving away from the decorator scheme which feels obstrusive sometimes.
    """

    def __init__(self, view_func: Callable, main_http_verb: str, rule: str):
        self.main_http_verb = main_http_verb
        self.default_param_location = self._solve_default_params_location(
            main_http_verb
        )
        self.rule = rule
        self.path_param_names = set(re.findall(r"<(?:\w*:)?(\w*?)>", rule))
        self.query_params: list[SolvedArgument] = []
        self.path_params: list[SolvedArgument] = []
        self.header_params: list[SolvedArgument] = []
        self.cookie_params: list[SolvedArgument] = []
        self.body_params: list[SolvedArgument] = []
        self.form_params: list[SolvedArgument] = []
        self.file_params: list[SolvedArgument] = []
        self.other_params: list[SolvedArgument] = []
        self.locations_to_visit: set[ArgumentLocation] = set()
        self._solve_params(view_func)
        self._check_compliance()
        self._body_field: SolvedArgument | None = None

    @staticmethod
    def _solve_default_params_location(
        main_http_verb: str,
    ) -> ArgumentLocation:
        """Return the default FieldInfo for the InboundHandler."""
        if main_http_verb in {"POST", "PUT"}:
            return ArgumentLocation.body
        if main_http_verb == "GET":
            return ArgumentLocation.query
        return ArgumentLocation.path

    @property
    def parameters(self) -> list[SolvedArgument]:
        """Return all Parameters of the InboundHandler."""
        return (
            self.query_params
            + self.path_params
            + self.header_params
            + self.cookie_params
        )

    @property
    def body_arguments(self) -> list[SolvedArgument]:
        """Return all Body Arguments of the InboundHandler."""
        return self.body_params + self.form_params + self.file_params

    @property
    def is_valid(self) -> bool:
        """Check if the InboundHandler has any Configured Parameters."""
        return len(self.locations_to_visit) > 0

    @property
    def has_request_body(self) -> bool:
        """Check if the InboundHandler has any Configured Parameters."""
        return len(self.body_arguments) > 0

    def body_field(self, name: str | None = None) -> SolvedArgument | None:
        """The Body Arguments are combined into a single Body Field."""
        if self._body_field is None and self.body_arguments and name is not None:
            self._body_field = self._build_body_field(name)
        return self._body_field

    def _solve_body_field_info(self) -> BodyArgument:
        """Return the Body FieldInfo for the InboundHandler.

        Has unreachable branches. Single Arguments never reach this method.
        """
        body_field_info_kwargs: dict[str, Any] = {"default": None, "embed": False}
        if len(self.file_params) > 0:  # pragma: no cover
            body_field_info: Callable = File
        elif len(self.form_params) > 0:
            body_field_info = Form
        else:
            body_field_info = Body
            body_param_media_types = [
                f.field_info.media_type
                for f in self.body_arguments
                if isinstance(f.field_info, BodyArgument)
            ]
            if len(set(body_param_media_types)) == 1:  # pragma: no cover
                body_field_info_kwargs["media_type"] = body_param_media_types[0]
        return body_field_info(**body_field_info_kwargs)

    def _build_body_field(self, name: str) -> SolvedArgument | None:
        """Build the Body Field from the Body Arguments.

        TODO: refactor this method: inelegant.
        TODO: warn if multiple locations are used.
        TODO: embed should be set to True if multiple BaseModel are used.
        embed should only be used to build the body field.
        """
        first_param = self.body_arguments[0]
        field_info = first_param.field_info
        body_param_names_set = {param.name for param in self.body_arguments}
        if (
            len(body_param_names_set) == 1
            and getattr(field_info, "embed", None) is not None
        ):
            if not _lenient_issubclass(first_param.annotation, BaseModel):
                first_param.embed = True
            return first_param
        model_name = f"{name}_request_body_as_model"
        any_embed = any(argument.embed for argument in self.body_arguments)
        any_required = any(argument.required for argument in self.body_arguments)
        field_defs: dict[str, Any] = {}
        for argument in self.body_arguments:
            argument.embed = any_embed
            if argument.required:
                field_defs[argument.name] = (argument.annotation, ...)
            else:
                field_defs[argument.name] = (argument.annotation, argument.default)
        BodyModel: type[BaseModel] = create_model(model_name, **field_defs)  # noqa: N806
        body_field_info = self._solve_body_field_info()
        return SolvedArgument.specialize(
            name=name,
            annotation=BodyModel,
            required=any_required,
            view_param=body_field_info,
        )

    def add_inbound_handling_to(
        self, view_func: JeroboamRouteCallable
    ) -> JeroboamRouteCallable:
        """It injects inbound parsed and validated data into the view function."""

        @wraps(view_func)
        def wrapper(*args, **kwargs) -> JeroboamResponseReturnValue:
            inbound_values, errors = self._parse_and_validate_inbound_data(**kwargs)
            if errors:
                raise InvalidRequest(errors)
            return view_func(*args, **inbound_values)

        return wrapper

    def _check_compliance(self):
        """Will warn the user if their view function does something a bit off."""
        if len(self.form_params + self.file_params) > 0 and self.main_http_verb not in {
            "POST",
            "PUT",
            "PATCH",
        }:
            import warnings

            warnings.warn(
                f"You have defined Form or File Parameters on a {self.main_http_verb} "
                "request. This is not supported by Flask: "
                "https://flask.palletsprojects.com/en/2.2.x/api/#incoming-request-data",
                UserWarning,
                stacklevel=2,
            )

    def _solve_params(self, view_func: Callable):
        """Registering the Parameters of the View Function."""
        signature = get_typed_signature(view_func)
        for parameter_name, parameter in signature.parameters.items():
            solved_param = self._solve_view_function_parameter(
                param_name=parameter_name, param=parameter
            )
            # Check if Param is in Path (not needed for now)
            self._register_view_parameter(solved_param)

    def _solve_view_function_parameter(
        self,
        param_name: str,
        param: inspect.Parameter,
        force_location: ArgumentLocation | None = None,
        ignore_default: bool = False,
    ) -> SolvedArgument:
        """Analyse the param and its annotation to solve its configiration.

        At the end of this process, we want to know the following things:
        - What is its location?
        - What is its type/annotation?
        - Is it a scalar or a sequence?
        - Is it required and/or has a default value?
        # Split it into functions for each step.
        """
        solved_location = self._solve_location(param_name, param, force_location)
        # Get the ViewParam
        if isinstance(param.default, ViewArgument):
            view_param = param.default
        else:
            param_class = get_argument_class(solved_location)
            raw_default = param.default
            if raw_default is param.empty or raw_default is Ellipsis:
                raw_default = PydanticUndefined
            view_param = param_class(raw_default)

        default_value = self._solve_default_value(param, ignore_default)
        # Solving Required
        required: bool = default_value is PydanticUndefined
        annotation = param.annotation if param.annotation != param.empty else Any

        return SolvedArgument.specialize(
            name=param_name,
            annotation=annotation,
            required=required,
            view_param=view_param,
        )

    def _solve_location(
        self,
        param_name: str,
        param: inspect.Parameter,
        force_location: ArgumentLocation | None = None,
    ) -> ArgumentLocation:
        if param_name in self.path_param_names:
            return ArgumentLocation.path
        return getattr(
            param.default, "location", force_location or self.default_param_location
        )

    def _solve_default_value(
        self,
        param: inspect.Parameter,
        ignore_default: bool,
    ) -> Any:
        default_value: Any = getattr(param.default, "default", param.default)
        if (
            default_value in {param.empty, Ellipsis, PydanticUndefined}
            or ignore_default
        ):
            default_value = PydanticUndefined
        return default_value

    def _register_view_parameter(self, solved_parameter: SolvedArgument) -> None:
        """Registering the Solved View parameters for the View Function.

        The registration will put the params in the right list
        and add the location to the locations_to_visit set.
        """
        assert solved_parameter.location is not None  # noqa: S101
        self.locations_to_visit.add(solved_parameter.location)
        {
            ArgumentLocation.query: self.query_params,
            ArgumentLocation.path: self.path_params,
            ArgumentLocation.header: self.header_params,
            ArgumentLocation.body: self.body_params,
            ArgumentLocation.form: self.form_params,
            ArgumentLocation.cookie: self.cookie_params,
            ArgumentLocation.file: self.file_params,
        }.get(solved_parameter.location, self.other_params).append(solved_parameter)

    def _parse_and_validate_inbound_data(self, **kwargs) -> tuple[dict, list[dict]]:
        """Parse and Validate the request Inbound data."""
        errors = []
        values = {}
        for param in self.parameters:
            values_, errors_ = param.validate_request()
            errors.extend(errors_)
            values.update(values_)
        if body_field := self.body_field(self.rule):
            body_value, body_errors = body_field.validate_request()
            errors.extend(body_errors)
            if len(self.body_arguments) > 1:
                values.update(_unpack_body_values(body_value))
            else:
                values.update(body_value)
        return values, errors



"""Testing Examples from README.md file."""

from docs_src.readme.readme00 import app as app00
from docs_src.readme.readme01 import app as app01


def test_readme_00(app=app00):
    """Test the first code snippet from the README.md file."""
    client = app.test_client()
    response = client.get("/ping")
    assert response.status_code == 200
    assert response.data == b"pong"


def test_readme_01(app=app01):
    """Test the second code snippet from the README.md file."""
    client = app.test_client()
    response = client.get("/wines?page=1&perPage=2&search=Champagne")
    assert response.status_code == 200
    assert response.json == [
        {"cuvee": "Brut - Blanc de Blancs", "appellation": "Champagne"},
        {"cuvee": "Grande Cuvée - 170ème Edition", "appellation": "Champagne"},
    ]



"""Testing the outbound feature page of our Documentation."""

import pytest

from docs_src.features.outbound import app


@pytest.fixture()
def client():
    """A Test Client for the app."""
    return app.test_client()


def test_explicit_response_model(client):
    """Test the implicit location GET example."""
    response = client.get("/tasks/42")

    assert response.status_code == 200
    assert response.json == {
        "id": 42,
        "name": "Find the answer.",
        "description": "Just here to make a point.",
    }


def test_dictionnary(client):
    """Test the implicit location GET example."""
    response = client.get("/tasks/42/no_response_model")

    assert response.status_code == 200
    assert response.json == {"id": 42, "name": "I'm from the dictionary."}


def test_implicit_response_model(client):
    """Test the implicit location GET example."""
    response = client.get("/tasks/42/implicit_from_annotation")

    assert response.status_code == 200
    assert response.json == {
        "id": 42,
        "name": "Implicit from Annotation",
        "description": "Just here to make a point.",
    }


def test_implicit_response_model_no_annotation(client):
    """Test the implicit location GET example."""
    response = client.get("/tasks/42/implicit_no_annotation")

    assert response.status_code == 500
    assert response.json == {
        "message": "Internal Error",
    }


def test_response_model_turned_off(client):
    """Test the implicit location GET example."""
    response = client.get("/tasks/42/response_model_off")

    assert response.status_code == 200
    assert response.json == {"id": 42, "name": "Response Model is off."}



"""Testing Error Handling."""

from flask.testing import FlaskClient

from flask_jeroboam.jeroboam import Jeroboam


def test_invalid_request(
    client: FlaskClient,
):
    """GIVEN a view function
    WHEN it raises a ResponseValidationError
    THEN I get a 400 response with message details
    """
    response = client.get("/invalid_request")
    assert response.status_code == 400
    assert response.json == {
        "detail": [
            {
                "loc": ["query", "missing_param"],
                "msg": "Field required",
                "type": "missing",
            }
        ]
    }


def test_ressource_not_found(
    client: FlaskClient,
):
    """GIVEN a decorated view function
    WHEN it raises a Generic ResssourceNotFound
    THEN I get a 404 response with RessourceNotFound Generic Message
    """
    response = client.get("/ressource_not_found")
    assert response.status_code == 404
    assert response.data.startswith(b"RessourceNotFound: TestRessource not found :")


def test_generic_ressource_not_found_message(
    client: FlaskClient,
):
    """GIVEN a decorated view function
    WHEN it raises a ResssourceNotFound
    THEN I get a 404 response with RessourceNotFound Generic Message
    """
    response = client.get("/generic_ressource")
    assert response.status_code == 404
    assert response.data.startswith(b"RessourceNotFound: My Message")


def test_internal_server_error(
    client: FlaskClient,
):
    """GIVEN a decorated view function
    WHEN it raises a ServerError
    THEN I get a 500 response with ServerError Message
    """
    response = client.get("/server_error")
    assert response.status_code == 500
    assert response.data.startswith(b"InternalServerError: My Message")


def test_opt_out_registering():
    """GIVEN a decorated view function
    WHEN it raises a ServerError
    THEN I get a 500 response with ServerError Message
    """
    app = Jeroboam(__name__)
    app.config["JEROBOAM_REGISTER_ERROR_HANDLERS"] = False
    app.init_app()

    @app.route("/will_raise_an_error")
    def server_error(failing_argument: int):
        return {"message": "Will not Return"}

    client = app.test_client()
    response = client.get("/will_raise_an_error?failing_argument=not_an_int")
    assert response.status_code == 400
    assert response.data.startswith(
        b"<!doctype html>\n<html lang=en>\n<title>400 Bad Request"
    )
